@@ -202,6 +202,7 @@ class GPUArray(object):
         self.strides = strides
         self.mem_size = self.size = s
         self.nbytes = self.dtype.itemsize * self.size
+        self.itemsize = self.dtype.itemsize
 
         self.allocator = allocator
         if gpudata is None:
@@ -213,7 +214,6 @@ class GPUArray(object):
             assert base is None
         else:
             self.gpudata = gpudata
-
         self.base = base
 
         self._grid, self._block = splay(self.mem_size)
@@ -275,7 +275,7 @@ class GPUArray(object):
         return self.get(ary=ary, async=True, stream=stream)
 
     def copy(self):
-        new = GPUArray(self.shape, self.dtype)
+        new = GPUArray(self.shape, self.dtype, self.allocator)
         _memcpy_discontig(new, self)
         return new
 
@@ -375,7 +375,7 @@ class GPUArray(object):
 
         return out
 
-    def _new_like_me(self, dtype=None):
+    def _new_like_me(self, dtype=None, order="C"):
         strides = None
         if dtype is None:
             dtype = self.dtype
@@ -384,7 +384,7 @@ class GPUArray(object):
                 strides = self.strides
 
         return self.__class__(self.shape, dtype,
-                allocator=self.allocator, strides=strides)
+                allocator=self.allocator, strides=strides, order=order)
 
     # operators ---------------------------------------------------------------
     def mul_add(self, selffac, other, otherfac, add_timer=None, stream=None):
@@ -680,9 +680,12 @@ class GPUArray(object):
 
         return result
 
-    def reshape(self, *shape):
+    def reshape(self, *shape, **kwargs):
         """Gives a new shape to an array without changing its data."""
 
+        # Python 2.x compatibility: use kwargs instead of named 'order' keyword
+        order = kwargs.pop("order", "C")
+
         # TODO: add more error-checking, perhaps
         if not self.flags.forc:
             raise RuntimeError("only contiguous arrays may "
@@ -691,7 +694,10 @@ class GPUArray(object):
         if isinstance(shape[0], tuple) or isinstance(shape[0], list):
             shape = tuple(shape[0])
 
-        if shape == self.shape:
+        same_contiguity = ((order == "C" and self.flags.c_contiguous) or
+                           (order == "F" and self.flags.f_contiguous))
+
+        if shape == self.shape and same_contiguity:
             return self
 
         if -1 in shape:
@@ -712,7 +718,8 @@ class GPUArray(object):
                 dtype=self.dtype,
                 allocator=self.allocator,
                 base=self,
-                gpudata=int(self.gpudata))
+                gpudata=int(self.gpudata),
+                order=order)
 
     def ravel(self):
         return self.reshape(self.size)
@@ -900,8 +907,11 @@ class GPUArray(object):
         if issubclass(dtype.type, np.complexfloating):
             from pytools import match_precision
             real_dtype = match_precision(np.dtype(np.float64), dtype)
-
-            result = self._new_like_me(dtype=real_dtype)
+            if self.flags.f_contiguous:
+                order = "F"
+            else:
+                order = "C"
+            result = self._new_like_me(dtype=real_dtype, order=order)
 
             func = elementwise.get_real_kernel(dtype, real_dtype)
             func.prepared_async_call(self._grid, self._block, None,
@@ -922,8 +932,11 @@ class GPUArray(object):
 
             from pytools import match_precision
             real_dtype = match_precision(np.dtype(np.float64), dtype)
-
-            result = self._new_like_me(dtype=real_dtype)
+            if self.flags.f_contiguous:
+                order = "F"
+            else:
+                order = "C"
+            result = self._new_like_me(dtype=real_dtype, order=order)
 
             func = elementwise.get_imag_kernel(dtype, real_dtype)
             func.prepared_async_call(self._grid, self._block, None,
@@ -941,7 +954,11 @@ class GPUArray(object):
                 raise RuntimeError("only contiguous arrays may "
                         "be used as arguments to this operation")
 
-            result = self._new_like_me()
+            if self.flags.f_contiguous:
+                order = "F"
+            else:
+                order = "C"
+            result = self._new_like_me(order=order)
 
             func = elementwise.get_conj_kernel(dtype)
             func.prepared_async_call(self._grid, self._block, None,
@@ -989,27 +1006,68 @@ empty = GPUArray
 
 def zeros(shape, dtype, allocator=drv.mem_alloc, order="C"):
     """Returns an array of the given shape and dtype filled with 0's."""
-
     result = GPUArray(shape, dtype, allocator, order=order)
     zero = np.zeros((), dtype)
     result.fill(zero)
     return result
 
 
-def empty_like(other_ary):
+def _array_like_helper(other_ary, dtype, order):
+    """Set order, strides, dtype as in numpy's zero_like. """
+    strides = None
+    if order == "A":
+        if other_ary.flags.f_contiguous and not other_ary.flags.c_contiguous:
+            order = "F"
+        else:
+            order = "C"
+    elif order == "K":
+        if other_ary.flags.c_contiguous or (other_ary.ndim <= 1):
+            order = "C"
+        elif other_ary.flags.f_contiguous:
+            order = "F"
+        else:
+            # array_like routines only return positive strides
+            strides = [np.abs(s) for s in other_ary.strides]
+            if dtype is not None and dtype != other_ary.dtype:
+                # scale strides by itemsize when dtype is not the same
+                itemsize = other_ary.nbytes // other_ary.size
+                itemsize_ratio = np.dtype(dtype).itemsize / itemsize
+                strides = [int(s*itemsize_ratio) for s in strides]
+    elif order not in ["C", "F"]:
+        raise ValueError("Unsupported order: %r" % order)
+    if dtype is None:
+        dtype = other_ary.dtype
+    return dtype, order, strides
+
+
+def empty_like(other_ary, dtype=None, order="K"):
+    dtype, order, strides = _array_like_helper(other_ary, dtype, order)
     result = GPUArray(
-            other_ary.shape, other_ary.dtype, other_ary.allocator)
+            other_ary.shape, dtype, other_ary.allocator, order=order,
+            strides=strides)
     return result
 
 
-def zeros_like(other_ary):
+def zeros_like(other_ary, dtype=None, order="K"):
+    dtype, order, strides = _array_like_helper(other_ary, dtype, order)
     result = GPUArray(
-            other_ary.shape, other_ary.dtype, other_ary.allocator)
+            other_ary.shape, dtype, other_ary.allocator, order=order,
+            strides=strides)
     zero = np.zeros((), result.dtype)
     result.fill(zero)
     return result
 
 
+def ones_like(other_ary, dtype=None, order="K"):
+    dtype, order, strides = _array_like_helper(other_ary, dtype, order)
+    result = GPUArray(
+            other_ary.shape, dtype, other_ary.allocator, order=order,
+            strides=strides)
+    one = np.ones((), result.dtype)
+    result.fill(one)
+    return result
+
+
 def arange(*args, **kwargs):
     """Create an array filled with numbers spaced `step` apart,
     starting from `start` and ending at `stop`.
@@ -1103,7 +1161,7 @@ def _compact_strides(a):
     # Compute strides to have same order as self, but packed
     info = sorted(
             (a.strides[axis], a.shape[axis], axis)
-            for axis in xrange(len(a.shape)))
+            for axis in range(len(a.shape)))
 
     strides = [None]*len(a.shape)
     stride = a.dtype.itemsize
@@ -1143,7 +1201,7 @@ def _memcpy_discontig(dst, src, async=False, stream=None):
         # and remove singleton axes
         src_info = sorted(
                 (src.strides[axis], axis)
-                for axis in xrange(len(src.shape)) if src.shape[axis] > 1)
+                for axis in range(len(src.shape)) if src.shape[axis] > 1)
         axes = [axis for _, axis in src_info]
         shape = [src.shape[axis] for axis in axes]
         src_strides = [src.strides[axis] for axis in axes]
