@@ -602,6 +602,9 @@ namespace pycuda
       CUcontext handle() const
       { return m_context; }
 
+      intptr_t handle_int() const
+      { return (intptr_t) m_context; }
+
       bool operator==(const context &other) const
       {
         return m_context == other.m_context;
@@ -1499,7 +1502,7 @@ namespace pycuda
             py::handle<>(
 #if PY_VERSION_HEX >= 0x03030000
               PyMemoryView_FromMemory((char *) (get_pointer() + offset), size,
-                PyBUF_READ | PyBUF_WRITE)
+                PyBUF_WRITE)
 #else /* Py2 */
               PyBuffer_FromReadWriteMemory((void *) (get_pointer() + offset), size)
 #endif
