@@ -354,6 +354,50 @@ class TestDriver:
             assert np.sum(np.abs(A_gpu.get()-np.transpose(A_cpu))) == np.array(0,dtype=prec)
             A_gpu.gpudata.free()
 
+    @mark_cuda_test
+    def test_2d_fp_texturesLayered(self):
+        orden = "F"
+        npoints = 32
+
+        for prec in [np.int16,np.float32,np.float64,np.complex64,np.complex128]:
+            prec_str = dtype_to_ctype(prec)
+            if prec == np.complex64: fpName_str = 'fp_tex_cfloat'
+            elif prec == np.complex128: fpName_str = 'fp_tex_cdouble'
+            elif prec == np.float64: fpName_str = 'fp_tex_double'
+            else: fpName_str = prec_str
+            A_cpu = np.zeros([npoints,npoints],order=orden,dtype=prec)
+            A_cpu[:] = np.random.rand(npoints,npoints)[:]
+            A_gpu = gpuarray.zeros(A_cpu.shape,dtype=prec,order=orden)
+
+            myKern = '''
+            #include <pycuda-helpers.hpp>
+            texture<fpName, cudaTextureType2DLayered, cudaReadModeElementType> mtx_tex;
+
+            __global__ void copy_texture(cuPres *dest)
+            {
+              int row = blockIdx.x*blockDim.x + threadIdx.x;
+              int col = blockIdx.y*blockDim.y + threadIdx.y;
+
+              dest[row + col*blockDim.x*gridDim.x] = fp_tex2DLayered(mtx_tex, col, row, 1);
+            }
+            '''
+            myKern = myKern.replace('fpName',fpName_str)
+            myKern = myKern.replace('cuPres',prec_str)
+            mod = SourceModule(myKern)
+
+            copy_texture = mod.get_function("copy_texture")
+            mtx_tex = mod.get_texref("mtx_tex")
+            cuBlock = (16,16,1)
+            if cuBlock[0]>npoints:
+                cuBlock = (npoints,npoints,1)
+            cuGrid   = (npoints//cuBlock[0]+1*(npoints % cuBlock[0] != 0 ),npoints//cuBlock[1]+1*(npoints % cuBlock[1] != 0 ),1)
+            copy_texture.prepare('P',texrefs=[mtx_tex])
+            cudaArray = drv.np_to_array(A_cpu,orden,allowSurfaceBind=True)
+            mtx_tex.set_array(cudaArray)
+            copy_texture.prepared_call(cuGrid,cuBlock,A_gpu.gpudata)
+            assert np.sum(np.abs(A_gpu.get()-np.transpose(A_cpu))) == np.array(0,dtype=prec)
+            A_gpu.gpudata.free()
+
     @mark_cuda_test
     def test_3d_fp_textures(self):
         orden = "C"
@@ -400,17 +444,30 @@ class TestDriver:
 
     @mark_cuda_test
     def test_3d_fp_surfaces(self):
-        orden = "F"
+        orden = "C"
         npoints = 32
 
         for prec in [np.int16,np.float32,np.float64,np.complex64,np.complex128]:
             prec_str = dtype_to_ctype(prec)
-            if prec == np.complex64: fpName_str = 'fp_tex_cfloat'
-            elif prec == np.complex128: fpName_str = 'fp_tex_cdouble'
-            elif prec == np.float64: fpName_str = 'fp_tex_double'
-            else: fpName_str = prec_str
-            A_cpu = np.zeros([npoints,npoints,npoints],order=orden,dtype=prec)
-            A_cpu[:] = np.random.rand(npoints,npoints,npoints)[:]
+            if prec == np.complex64:
+                fpName_str = 'fp_tex_cfloat'
+                A_cpu = np.zeros([npoints,npoints,npoints],order=orden,dtype=prec)
+                A_cpu[:].real = np.random.rand(npoints,npoints,npoints)[:]
+                A_cpu[:].imag = np.random.rand(npoints,npoints,npoints)[:]
+            elif prec == np.complex128:
+                fpName_str = 'fp_tex_cdouble'
+                A_cpu = np.zeros([npoints,npoints,npoints],order=orden,dtype=prec)
+                A_cpu[:].real = np.random.rand(npoints,npoints,npoints)[:]
+                A_cpu[:].imag = np.random.rand(npoints,npoints,npoints)[:]
+            elif prec == np.float64:
+                fpName_str = 'fp_tex_double'
+                A_cpu = np.zeros([npoints,npoints,npoints],order=orden,dtype=prec)
+                A_cpu[:] = np.random.rand(npoints,npoints,npoints)[:]
+            else:
+                fpName_str = prec_str
+                A_cpu = np.zeros([npoints,npoints,npoints],order=orden,dtype=prec)
+                A_cpu[:] = np.random.rand(npoints,npoints,npoints)[:]*100.
+
             A_gpu = gpuarray.to_gpu(A_cpu) # Array randomized
 
             myKernRW = '''
@@ -567,7 +624,7 @@ class TestDriver:
 
         for e in range(e0-6, e0-4):
             for i in range(100):
-                queue.append(pool.allocate(1<<e))
+                queue.append(pool.allocate(1 << e))
                 if len(queue) > 10:
                     queue.pop(0)
         del queue
@@ -575,9 +632,9 @@ class TestDriver:
 
     @mark_cuda_test
     def test_multi_context(self):
-        if drv.get_version() < (2,0,0):
+        if drv.get_version() < (2, 0, 0):
             return
-        if drv.get_version() >= (2,2,0):
+        if drv.get_version() >= (2, 2, 0) and drv.get_version() < (8,):
             if drv.Context.get_device().compute_mode == drv.compute_mode.EXCLUSIVE:
                 return
 
@@ -773,10 +830,7 @@ class TestDriver:
             from py.test import skip
             skip("register_host_memory is not supported on OS X")
 
-        import resource
-
-        a = drv.aligned_empty((2**20,), np.float64,
-            alignment=resource.getpagesize())
+        a = drv.aligned_empty((2**20,), np.float64)
         a_pin = drv.register_host_memory(a)
 
         gpu_ary = drv.mem_alloc_like(a)
@@ -784,7 +838,6 @@ class TestDriver:
         drv.memcpy_htod_async(gpu_ary, a_pin, stream)
         drv.Context.synchronize()
 
-    @pytest.mark.xfail
     @mark_cuda_test
     # https://github.com/inducer/pycuda/issues/45
     def test_recursive_launch(self):
@@ -840,9 +893,8 @@ class TestDriver:
             drv.memcpy_htod(a_gpu, a)
             drv.memcpy_htod(b_gpu, b)
 
-            mod = SourceModule(cuda_string,
-                    options=['-rdc=true', '-lcudadevrt'],
-                    keep=True)
+            from pycuda.compiler import DynamicSourceModule
+            mod = DynamicSourceModule(cuda_string, keep=True)
 
             func = mod.get_function("math")
             func(a_gpu, b_gpu, c_gpu, d_gpu, e_gpu, f_gpu,
@@ -864,6 +916,34 @@ class TestDriver:
 
         math(a, b, c, d, e, f)
 
+    @mark_cuda_test
+    def test_jit_link_module(self):
+        if drv.Context.get_device().compute_capability() < (3, 5):
+            from pytest import skip
+            skip("need compute capability 3.5 or higher for dynamic parallelism")
+
+        test_outer_cu = '''#include <cstdio>
+        __global__ void test_kernel() {
+            extern __global__ void test_kernel_inner();
+            printf("Hello outer world!\\n");
+            test_kernel_inner<<<2, 1>>>();
+        }'''
+
+        test_inner_cu = '''#include <cstdio>
+        __global__ void test_kernel_inner() {
+            printf("  Hello inner world!\\n");
+        }'''
+
+        from pycuda.compiler import DynamicModule
+        mod = DynamicModule()
+        mod.add_source(test_outer_cu, nvcc_options=['-rdc=true', '-lcudadevrt'])
+        mod.add_source(test_inner_cu, nvcc_options=['-rdc=true', '-lcudadevrt'])
+        mod.add_stdlib('cudadevrt')
+        mod.link()
+
+        test_kernel = mod.get_function('test_kernel')
+        test_kernel(grid=(2,1), block=(1,1,1))
+
 
 def test_import_pyopencl_before_pycuda():
     try:
