@@ -691,6 +691,25 @@ class TestGPUArray:
             assert la.norm(z.get().imag - z.imag.get()) == 0
             assert la.norm(z.get().conj() - z.conj().get()) == 0
 
+            # verify contiguity is preserved
+            for order in ["C", "F"]:
+                # test both zero and non-zero value code paths
+                z_real = gpuarray.zeros(z.shape, dtype=real_dtype,
+                                        order=order)
+                z2 = z.reshape(z.shape, order=order)
+                for zdata in [z_real, z2]:
+                    if order == "C":
+                        assert zdata.flags.c_contiguous == True
+                        assert zdata.real.flags.c_contiguous == True
+                        assert zdata.imag.flags.c_contiguous == True
+                        assert zdata.conj().flags.c_contiguous == True
+                    elif order == "F":
+                        assert zdata.flags.f_contiguous == True
+                        assert zdata.real.flags.f_contiguous == True
+                        assert zdata.imag.flags.f_contiguous == True
+                        assert zdata.conj().flags.f_contiguous == True
+
+
     @mark_cuda_test
     def test_pass_slice_to_kernel(self):
         mod = SourceModule("""
@@ -779,6 +798,17 @@ class TestGPUArray:
             throws_exception = True
         assert throws_exception
 
+        # with order specified
+        a_gpu = a_gpu.reshape((4, 32), order='C')
+        assert a_gpu.flags.c_contiguous
+        a_gpu = a_gpu.reshape(4, 32, order='F')
+        assert a_gpu.flags.f_contiguous
+        a_gpu = a_gpu.reshape((4, 32), order='F')
+        assert a_gpu.flags.f_contiguous
+        # default is C-contiguous
+        a_gpu = a_gpu.reshape((4, 32))
+        assert a_gpu.flags.c_contiguous
+
     @mark_cuda_test
     def test_view(self):
         a = np.arange(128).reshape(8, 16).astype(np.float32)
@@ -805,12 +835,12 @@ class TestGPUArray:
         # Slice with length 1 on dimensions 0 and 1
         a_gpu_slice = a_gpu[0:1,1:2,:,:]
         assert a_gpu_slice.shape == (1,1,shape[2],shape[3])
-        assert a_gpu_slice.flags.c_contiguous is False 
+        assert a_gpu_slice.flags.c_contiguous
 
         # Squeeze it and obtain contiguity
         a_gpu_squeezed_slice = a_gpu[0:1,1:2,:,:].squeeze()
         assert a_gpu_squeezed_slice.shape == (shape[2],shape[3])
-        assert a_gpu_squeezed_slice.flags.c_contiguous is True 
+        assert a_gpu_squeezed_slice.flags.c_contiguous
 
         # Check that we get the original values out
         assert np.all(a_gpu_slice.get().ravel() == a_gpu_squeezed_slice.get().ravel())
@@ -818,12 +848,12 @@ class TestGPUArray:
         # Slice with length 1 on dimensions 2
         a_gpu_slice = a_gpu[:,:,2:3,:]
         assert a_gpu_slice.shape == (shape[0],shape[1],1,shape[3])
-        assert a_gpu_slice.flags.c_contiguous is False
+        assert not a_gpu_slice.flags.c_contiguous
 
         # Squeeze it, but no contiguity here
         a_gpu_squeezed_slice = a_gpu[:,:,2:3,:].squeeze()
         assert a_gpu_squeezed_slice.shape == (shape[0],shape[1],shape[3])
-        assert a_gpu_squeezed_slice.flags.c_contiguous is False
+        assert not a_gpu_squeezed_slice.flags.c_contiguous
 
         # Check that we get the original values out
         assert np.all(a_gpu_slice.get().ravel() == a_gpu_squeezed_slice.get().ravel())
@@ -1055,6 +1085,56 @@ class TestGPUArray:
         assert np.allclose(a_gpu.get(), a)
         assert np.allclose(a_gpu[1:3,1:3,1:3].get(), a[1:3,1:3,1:3])
 
+    @mark_cuda_test
+    def test_zeros_like_etc(self):
+        shape = (16, 16)
+        a = np.random.randn(*shape).astype(np.float32)
+        z = gpuarray.to_gpu(a)
+        zf = gpuarray.to_gpu(np.asfortranarray(a))
+        a_noncontig = np.arange(3*4*5).reshape(3, 4, 5).swapaxes(1, 2)
+        z_noncontig = gpuarray.to_gpu(a_noncontig)
+        for func in [gpuarray.empty_like,
+                     gpuarray.zeros_like,
+                     gpuarray.ones_like]:
+            for arr in [z, zf, z_noncontig]:
+
+                contig = arr.flags.c_contiguous or arr.flags.f_contiguous
+
+                # Output matches order of input.
+                # Non-contiguous becomes C-contiguous
+                new_z = func(arr, order="A")
+                if contig:
+                    assert new_z.flags.c_contiguous == arr.flags.c_contiguous
+                    assert new_z.flags.f_contiguous == arr.flags.f_contiguous
+                else:
+                    assert new_z.flags.c_contiguous is True
+                    assert new_z.flags.f_contiguous is False
+                assert new_z.dtype == arr.dtype
+                assert new_z.shape == arr.shape
+
+                # Force C-ordered output
+                new_z = func(arr, order="C")
+                assert new_z.flags.c_contiguous is True
+                assert new_z.flags.f_contiguous is False
+                assert new_z.dtype == arr.dtype
+                assert new_z.shape == arr.shape
+
+                # Force Fortran-orded output
+                new_z = func(arr, order="F")
+                assert new_z.flags.c_contiguous is False
+                assert new_z.flags.f_contiguous is True
+                assert new_z.dtype == arr.dtype
+                assert new_z.shape == arr.shape
+
+                # Change the dtype, but otherwise match order & strides
+                # order = "K" so non-contiguous array remains non-contiguous
+                new_z = func(arr, dtype=np.complex64, order="K")
+                assert new_z.flags.c_contiguous == arr.flags.c_contiguous
+                assert new_z.flags.f_contiguous == arr.flags.f_contiguous
+                assert new_z.dtype == np.complex64
+                assert new_z.shape == arr.shape
+
+
 if __name__ == "__main__":
     # make sure that import failures get reported, instead of skipping the tests.
     import pycuda.autoinit  # noqa
